@@ -256,8 +256,8 @@ def lower_xla_callable(fun: lu.WrappedFun, device, backend, name,
   module_name = f"jit_{fun.__name__}"
   if config.jax_enable_mlir:
     module = mlir.lower_jaxpr_to_module(
-        module_name, closed_jaxpr, backend.platform, axis_env, name_stack,
-        donated_invars)
+        module_name, closed_jaxpr, backend.platform,
+        mlir.ReplicaAxisContext(axis_env), name_stack, donated_invars)
   else:
     module = xla.lower_jaxpr_to_xla_module(
         module_name, closed_jaxpr, backend.platform, axis_env,
@@ -557,7 +557,8 @@ def compile_or_get_cached(backend, computation, compile_options):
   from jax.experimental.compilation_cache import compilation_cache as cc
 
   if isinstance(computation, ir.Module):
-    module_name = computation.operation.name
+    sym_name = computation.operation.attributes['sym_name']
+    module_name = ir.StringAttr(sym_name).value
     computation = mlir.module_to_string(computation)
   else:
     module_name = computation.name()
