@@ -16,9 +16,10 @@
 # checking on import.
 
 import platform
+import re
 import os
 import warnings
-from typing import Optional
+from typing import Optional, Tuple
 
 __all__ = [
   'cuda_linalg', 'cuda_prng', 'cusolver', 'rocsolver', 'jaxlib', 'lapack',
@@ -26,8 +27,8 @@ __all__ = [
   'xla_extension',
 ]
 
-# First, before attempting to from jax import jaxlib, warn about experimental machine
-# configurations.
+# Before attempting to import jaxlib, warn about experimental
+# machine configurations.
 if platform.system() == "Darwin" and platform.machine() == "arm64":
   warnings.warn("JAX on Mac ARM machines is experimental and minimally tested. "
                 "Please see https://github.com/google/jax/issues/5501 in the "
@@ -41,33 +42,58 @@ except ModuleNotFoundError as err:
     'https://github.com/google/jax#installation for installation instructions.'
     ) from err
 
+import jax.version
 from jax.version import _minimum_jaxlib_version as _minimum_jaxlib_version_str
 try:
-  import jaxlib.version as jaxlib_version
+  import jaxlib.version
 except Exception as err:
   # jaxlib is too old to have version number.
   msg = f'This version of jax requires jaxlib version >= {_minimum_jaxlib_version_str}.'
   raise ImportError(msg) from err
 
-version = tuple(int(x) for x in jaxlib_version.__version__.split('.'))
-_minimum_jaxlib_version = tuple(int(x) for x in _minimum_jaxlib_version_str.split('.'))
 
-# Check the jaxlib version before importing anything else from jaxlib.
-def _check_jaxlib_version():
-  if version < _minimum_jaxlib_version:
-    msg = (f'jaxlib is version {jaxlib_version.__version__}, '
-           f'but this version of jax requires version {_minimum_jaxlib_version_str}.')
+# Checks the jaxlib version before importing anything else from jaxlib.
+# Returns the jaxlib version string.
+def check_jaxlib_version(jax_version: str, jaxlib_version: str,
+                         minimum_jaxlib_version: str):
+  # Regex to match a dotted version prefix 0.1.23.456.789 of a PEP440 version.
+  # PEP440 allows a number of non-numeric suffixes, which we allow also.
+  # We currently do not allow an epoch.
+  version_regex = re.compile(r"[0-9]+(?:\.[0-9]+)*")
+  def _parse_version(v: str) -> Tuple[int, ...]:
+    m = version_regex.match(v)
+    if m is None:
+      raise ValueError(f"Unable to parse jaxlib version '{v}'")
+    return tuple(int(x) for x in m.group(0).split('.'))
 
-    if version == (0, 1, 23):
-      msg += ('\n\nA common cause of this error is that you installed jaxlib '
-              'using pip, but your version of pip is too old to support '
-              'manylinux2010 wheels. Try running:\n\n'
-              'pip install --upgrade pip\n'
-              'pip install --upgrade jax jaxlib\n')
-    raise ValueError(msg)
+  _jax_version = _parse_version(jax_version)
+  _minimum_jaxlib_version = _parse_version(minimum_jaxlib_version)
+  _jaxlib_version = _parse_version(jaxlib_version)
 
-_check_jaxlib_version()
+  if _jaxlib_version < _minimum_jaxlib_version:
+    msg = (f'jaxlib is version {jaxlib_version}, but this version '
+           f'of jax requires version >= {minimum_jaxlib_version}.')
+    raise RuntimeError(msg)
 
+  if _jaxlib_version > _jax_version:
+    msg = (f'jaxlib version {jaxlib_version} is newer than and '
+           f'incompatible with jax version {jax_version}. Please '
+           'update your jax and/or jaxlib packages.')
+    raise RuntimeError(msg)
+
+  return _jaxlib_version
+
+version_str = jaxlib.version.__version__
+version = check_jaxlib_version(
+  jax_version=jax.version.__version__,
+  jaxlib_version=jaxlib.version.__version__,
+  minimum_jaxlib_version=jax.version._minimum_jaxlib_version)
+
+
+
+# Before importing any C compiled modules from jaxlib, first import the CPU
+# feature guard module to verify that jaxlib was compiled in a way that only
+# uses instructions that are present on this machine.
 import jaxlib.cpu_feature_guard as cpu_feature_guard
 cpu_feature_guard.check_cpu_features()
 
@@ -126,3 +152,6 @@ cuda_path: Optional[str]
 cuda_path = os.path.join(os.path.dirname(jaxlib.__file__), "cuda")
 if not os.path.isdir(cuda_path):
   cuda_path = None
+
+if xla_extension_version >= 58:
+  transfer_guard_lib = xla_client._xla.transfer_guard_lib
