@@ -458,12 +458,6 @@ def supported_dtypes():
     types -= {np.uint64, np.int64, np.float64, np.complex128}
   return types
 
-def skip_if_unsupported_type(dtype):
-  dtype = np.dtype(dtype)
-  if dtype.type not in supported_dtypes():
-    raise unittest.SkipTest(
-      f"Type {dtype.name} not supported on {device_under_test()}")
-
 def is_device_rocm():
   return xla_bridge.get_backend().platform_version.startswith('rocm')
 
@@ -615,7 +609,10 @@ def _rand_dtype(rand, shape, dtype, scale=1., post=lambda x: x):
     An ndarray of the given shape and dtype using random values based on a call
     to rand but scaled, converted to the appropriate dtype, and post-processed.
   """
-  r = lambda: np.asarray(scale * rand(*_dims_of_shape(shape)), dtype)
+  if _dtypes.issubdtype(dtype, np.unsignedinteger):
+    r = lambda: np.asarray(scale * abs(rand(*_dims_of_shape(shape))), dtype)
+  else:
+    r = lambda: np.asarray(scale * rand(*_dims_of_shape(shape)), dtype)
   if _dtypes.issubdtype(dtype, np.complexfloating):
     vals = r() + 1.0j * r()
   else:
@@ -918,7 +915,11 @@ def with_config(**kwds):
 
 class JaxTestCase(parameterized.TestCase):
   """Base class for JAX tests including numerical checks and boilerplate."""
-  _default_config = {'jax_enable_checks': True}
+  _default_config = {
+    'jax_enable_checks': True,
+    'jax_numpy_rank_promotion': 'raise',
+    'jax_traceback_filtering': 'off',
+  }
 
   # TODO(mattjj): this obscures the error messages from failures, figure out how
   # to re-enable it
@@ -1129,6 +1130,16 @@ def restore_spmd_lowering_flag():
   if old_spmd_lowering_flag is None: return
   config.update('experimental_xmap_spmd_lowering', old_spmd_lowering_flag)
 
+old_spmd_manual_lowering_flag = None
+def set_spmd_manual_lowering_flag(val: bool):
+  global old_spmd_manual_lowering_flag
+  old_spmd_manual_lowering_flag = config.experimental_xmap_spmd_lowering_manual
+  config.update('experimental_xmap_spmd_lowering_manual', val)
+
+def restore_spmd_manual_lowering_flag():
+  if old_spmd_manual_lowering_flag is None: return
+  config.update('experimental_xmap_spmd_lowering_manual', old_spmd_manual_lowering_flag)
+
 def create_global_mesh(mesh_shape, axis_names):
   size = prod(mesh_shape)
   if len(api.devices()) < size:
