@@ -87,6 +87,9 @@ class Jaxpr:
                    custom_pp_eqn_rules=custom_pp_eqn_rules)
     return doc.format(**kw)
 
+  def _repr_pretty_(self, p, cycle):
+    return p.text(self.pretty_print(use_color=True))
+
 
 def jaxprs_in_params(params) -> Iterator[Jaxpr]:
   for val in params.values():
@@ -142,6 +145,10 @@ class ClosedJaxpr:
     return pp_jaxpr(self.jaxpr, JaxprPpContext(), source_info=source_info,
                     print_shapes=print_shapes).format(**kw)
 
+
+  def _repr_pretty_(self, p, cycle):
+    return p.text(self.pretty_print(use_color=True))
+
 @curry
 def jaxpr_as_fun(closed_jaxpr: ClosedJaxpr, *args):
   return eval_jaxpr(closed_jaxpr.jaxpr, closed_jaxpr.consts, *args)
@@ -1331,9 +1338,12 @@ pytype_aval_mappings[Token] = lambda _: abstract_token
 
 
 def raise_to_shaped(aval: AbstractValue, weak_type=None):
+  aval_type = type(aval)
+  if aval_type is ShapedArray and weak_type is None:
+    return aval
   if weak_type is None:
     weak_type = getattr(aval, 'weak_type', False)
-  for typ in type(aval).__mro__:
+  for typ in aval_type.__mro__:
     handler = raise_to_shaped_mappings.get(typ)
     if handler: return handler(aval, weak_type)
   raise TypeError(type(aval))
@@ -1344,8 +1354,7 @@ raise_to_shaped_mappings : Dict[type, Callable] = {
   Bot: lambda aval, _: aval,
   UnshapedArray: lambda aval, _: aval,
   ShapedArray: lambda aval, weak_type: ShapedArray(
-      aval.shape, dtypes.canonicalize_dtype(aval.dtype), weak_type,
-      aval.named_shape)
+      aval.shape, aval.dtype, weak_type, aval.named_shape)
 }
 
 ### Operations on shapes and dimension sizes.
@@ -1561,7 +1570,7 @@ def canonicalize_shape(shape: Shape, context: str="") -> Shape:
     A tuple of canonical dimension values.
   """
   try:
-    return tuple(map(_canonicalize_dimension, shape))
+    return tuple(unsafe_map(_canonicalize_dimension, shape))
   except TypeError:
     pass
   raise _invalid_shape_error(shape, context)
@@ -1793,10 +1802,9 @@ def map_bind(primitive: 'MapPrimitive', fun, *args, out_axes_thunk, **params):
       out_axes = t(out_axes)
     return out_axes
   params = dict(params, out_axes_thunk=new_out_axes_thunk)
-  params_tuple = tuple(params.items())
   top_trace = find_top_trace(args)
   fun, todo_and_xforms = process_env_traces_map(
-      fun, primitive, top_trace and top_trace.level, params_tuple)
+      fun, primitive, top_trace and top_trace.level, tuple(params.items()))
   tracers = map(top_trace.full_raise, args)
   outs = primitive.process(top_trace, fun, tracers, params)
   env_trace_todo, _ = todo_and_xforms()
@@ -1824,14 +1832,16 @@ def process_env_traces_map(primitive: MapPrimitive, level: int,
   yield outs, (tuple(todo), tuple(out_axes_transforms))
 
 
-def mapped_aval(size: int, axis: int, aval: AbstractValue) -> AbstractValue:
+def mapped_aval(size: int, axis: Optional[int], aval: AbstractValue
+                ) -> AbstractValue:
   handler, _ = aval_mapping_handlers.get(type(aval), (None, None))
   if handler is not None:
     return handler(size, axis, aval)
   else:
     raise TypeError(f"no mapping handler for {aval} of type {type(aval)}")
 
-def unmapped_aval(size: int, axis_name, axis: int, aval: AbstractValue) -> AbstractValue:
+def unmapped_aval(size: int, axis_name, axis: Optional[int], aval: AbstractValue
+                  ) -> AbstractValue:
   _, handler = aval_mapping_handlers.get(type(aval), (None, None))
   if handler is not None:
     return handler(size, axis_name, axis, aval)
@@ -1841,16 +1851,20 @@ def unmapped_aval(size: int, axis_name, axis: int, aval: AbstractValue) -> Abstr
 def _map_unit(*_) -> AbstractUnit:
   return abstract_unit
 
-def _map_shaped_array(size: int, axis: int, aval: ShapedArray) -> ShapedArray:
-  assert aval.shape[axis] == size
+def _map_shaped_array(size: int, axis: Optional[int], aval: ShapedArray
+                      ) -> ShapedArray:
+  assert axis is None or aval.shape[axis] == size
   # TODO: Extend the named shape
+  if axis is None: return aval
   return ShapedArray(tuple_delete(aval.shape, axis), aval.dtype,
                      named_shape=aval.named_shape)
 
-def _unmap_shaped_array(size: int, axis_name, axis: int, aval: ShapedArray) -> ShapedArray:
+def _unmap_shaped_array(size: int, axis_name, axis: Optional[int],
+                        aval: ShapedArray) -> ShapedArray:
   named_shape = dict(aval.named_shape)
   # TODO: Make this mandatory
   named_shape.pop(axis_name, None)
+  if axis is None: return aval.replace(named_shape=named_shape)
   return ShapedArray(tuple_insert(aval.shape, axis, size), aval.dtype,
                      named_shape=named_shape)
 
@@ -2248,7 +2262,7 @@ def pp_vars(vs: Sequence[Any], context: JaxprPpContext,
     return pp.nest(2, pp.group(
       pp.join(pp.text(separator) + pp.group(pp.brk()), [
         pp.text(pp_var(v, context)) +
-        pp.dim(pp.text(":" + pp_aval(v.aval, context)))
+        pp.type_annotation(pp.text(":" + pp_aval(v.aval, context)))
         for v in vs
       ])
     ))
@@ -2304,13 +2318,18 @@ def pp_eqns(eqns, context: JaxprPpContext, *, print_shapes=True,
     [pp_eqn(e, context, print_shapes=print_shapes, source_info=source_info,
             custom_pp_eqn_rules=custom_pp_eqn_rules) for e in eqns])
 
-def pp_eqn_compact(primitive_name: str, params: Dict, context: JaxprPpContext
-                  ) -> pp.Doc:
-  filtered_params = {k: v for k, v in params.items()
-                     if (k != 'branches' and
-                         not isinstance(v, (Jaxpr, ClosedJaxpr)))}
-  return (pp.text(primitive_name) +
-          pp_kv_pairs(sorted(filtered_params.items()), context))
+def _compact_eqn_should_include(k: str, v: Any) -> bool:
+  if k == 'branches': return False
+  if isinstance(v, (Jaxpr, ClosedJaxpr)): return False
+  if (isinstance(v, tuple) and
+      any(isinstance(e, (Jaxpr, ClosedJaxpr)) for e in v)): return False
+  return True
+
+def str_eqn_compact(primitive_name: str, params: Dict) -> str:
+  "Compact equation to string conversion used in HLO metadata."
+  kvs = " ".join(f"{k}={v}" for k, v in params.items()
+                 if _compact_eqn_should_include(k, v))
+  return f"{primitive_name}[{kvs}]" if len(kvs) > 0 else primitive_name
 
 def pp_jaxpr_skeleton(jaxpr, eqns_fn, context: JaxprPpContext, *,
                       print_shapes=True) -> pp.Doc:
@@ -2321,11 +2340,11 @@ def pp_jaxpr_skeleton(jaxpr, eqns_fn, context: JaxprPpContext, *,
     pp.text("("), pp_vars(jaxpr.outvars, context, separator=","),
     pp.text(")" if len(jaxpr.outvars) != 1 else ",)")])
   return pp.group(pp.nest(2, pp.concat([
-    pp.text("{ "), pp.bright(pp.text("lambda ")),
+    pp.text("{ "), pp.keyword(pp.text("lambda ")),
     constvars, pp.text("; "), invars,
-    pp.text(". "), pp.bright(pp.text("let")),
+    pp.text(". "), pp.keyword(pp.text("let")),
     pp.nest(2, pp.brk() + eqns), pp.brk(),
-    pp.bright(pp.text("in ")), outvars
+    pp.keyword(pp.text("in ")), outvars
   ])) + pp.text(" }"))
 
 
