@@ -46,21 +46,21 @@ def _convert_list_args_to_tuple(f):
   return wrapper
 
 
-def _canonicalize_mesh_axes(mesh_axes):
+def _get_array_mapping(mesh_axes):
+  # Import here to avoid cyclic import error when importing gda in pjit.py.
+  from jax.experimental.pjit import get_array_mapping, _prepare_axis_resources
+
   if not isinstance(mesh_axes, PartitionSpec):
     pspec = PartitionSpec(*mesh_axes)
   else:
     pspec = mesh_axes
-  return pspec
+  parsed_pspec, _, _ = _prepare_axis_resources(pspec, "mesh_axes")
+  return get_array_mapping(parsed_pspec)
+
 
 def _get_indices(global_shape: Shape, global_mesh: pxla.Mesh,
                  mesh_axes: MeshAxes) -> Tuple[Index, ...]:
-  # Import here to avoid cyclic import error when importing gda in pjit.py.
-  from jax.experimental.pjit import get_array_mapping, _prepare_axis_resources
-
-  pspec = _canonicalize_mesh_axes(mesh_axes)
-  parsed_pspec, _, _ = _prepare_axis_resources(pspec, "mesh_axes")
-  array_mapping = get_array_mapping(parsed_pspec)
+  array_mapping = _get_array_mapping(mesh_axes)
   # The dtype doesn't matter for creating sharding specs.
   aval = core.ShapedArray(global_shape, np.float32)
   sharding_spec = pxla.mesh_sharding_specs(
@@ -93,11 +93,23 @@ def _get_shard_indices_replica_ids_uncached(
   indices = _get_indices(global_shape, global_mesh, mesh_axes)
   index_to_replica: Dict[int, int] = Counter()
   out = {}
+  unique_shards = 0
   for device, index in safe_zip(global_mesh.devices.flat, indices):
     h_index = _hashed_index(index)
     replica_id = index_to_replica[h_index]
+    if replica_id == 0:
+      unique_shards += 1
     index_to_replica[h_index] += 1
     out[device] = (index, replica_id)
+
+  shard_shape = get_shard_shape(global_shape, global_mesh, mesh_axes)
+  expected_unique_shards = prod(
+      [g // s for g, s in safe_zip(global_shape, shard_shape) if g != 0 or s != 0])
+  if expected_unique_shards != unique_shards:
+    raise RuntimeError(
+        f'Number of expected unique shards are: {expected_unique_shards} but '
+        f'got {unique_shards}. Please file a bug at '
+        'https://github.com/google/jax/issues.')
   return out
 
 
@@ -197,6 +209,8 @@ class GlobalDeviceArray:
   Attributes:
     shape : Global shape of the array.
     dtype : Dtype of the global array.
+    ndim : Number of array dimensions in the global shape.
+    size: Number of elements in the global array.
     local_shards : List of :class:`Shard` on the local devices of the current process.
       Data is materialized for all local shards.
     global_shards : List of all :class:`Shard` of the global array. Data isn’t
@@ -258,14 +272,20 @@ class GlobalDeviceArray:
       self._local_devices = self._global_mesh.local_devices
     else:
       self._local_devices = self._gda_fast_path_args.local_devices
-    assert len(device_buffers) == len(self._local_devices)
+
+    for db, ld in safe_zip(device_buffers, self._local_devices):
+      if db.device() != ld:
+        raise ValueError(
+            "The `global_mesh.local_devices` and `device_buffers` device order "
+            "doesn't match. Please use `global_mesh.local_devices` to put "
+            "arrays on devices instead of `jax.local_devices()`")
 
     self._local_shards = self._create_local_shards()
 
     ss = get_shard_shape(self._global_shape, self._global_mesh, self._mesh_axes)
     assert all(db.shape == ss for db in device_buffers), (
         f"Expected shard shape {ss} doesn't match the device buffer "
-        f"shape {device_buffers[0].shape}")
+        f"shape, got: {[db.shape for db in device_buffers]}")
 
     dtype = device_buffers[0].dtype
     assert all(db.dtype == dtype for db in device_buffers), (
@@ -294,6 +314,14 @@ class GlobalDeviceArray:
   def shape(self) -> Shape:
     return self._global_shape
 
+  @property
+  def ndim(self):
+    return len(self.shape)
+
+  @property
+  def size(self):
+    return prod(self.shape)
+
   @property
   def is_fully_replicated(self) -> bool:
     return self.shape == self.local_data(0).shape
@@ -312,8 +340,7 @@ class GlobalDeviceArray:
       out.append(Shard(device, index, rid, db))
     return out
 
-
-  @pxla.maybe_cached_property
+  @property
   def local_shards(self) -> Sequence[Shard]:
     for s in self._local_shards:
       # Ignore the type because mypy thinks data is None but local_shards
@@ -322,7 +349,7 @@ class GlobalDeviceArray:
         s.data.aval = core.ShapedArray(s.data.shape, s.data.dtype)  # type: ignore
     return self._local_shards
 
-  @pxla.maybe_cached_property
+  @property
   def global_shards(self) -> Sequence[Shard]:
     # Populating global_shards lazily (i.e. when requested) because populating
     # sthem eagerly leads to a performance regression when training on large
