@@ -15,7 +15,7 @@
 from enum import IntEnum
 import numpy as np
 from collections import OrderedDict, Counter
-from typing import Callable, Sequence, Tuple, Union
+from typing import Callable, Sequence, Tuple, Union, Optional
 from warnings import warn
 import itertools as it
 from functools import partial
@@ -40,7 +40,9 @@ from jax.interpreters import batching
 from jax.interpreters import partial_eval as pe
 from jax.interpreters.sharded_jit import PartitionSpec
 from jax._src.lib import xla_client as xc
-from jax.tree_util import tree_map, tree_flatten, tree_unflatten
+from jax.tree_util import (tree_map, tree_flatten, tree_unflatten,
+                           treedef_is_leaf, tree_structure)
+from jax._src.tree_util import prefix_errors
 from jax._src.util import (extend_name_stack, HashableFunction, safe_zip,
                          wrap_name, wraps, distributed_debug_log,
                          split_list, cache, tuple_insert)
@@ -207,7 +209,7 @@ def pjit(fun: Callable,
                        f"was called with only {len(args)} positional arguments.")
 
     # Putting this outside of wrapped would make resources lexically scoped
-    resource_env = maps.thread_resources.env
+    resource_env = pxla.thread_resources.env
     mesh = resource_env.physical_mesh
     if mesh.empty:
       raise RuntimeError("pjit requires a non-empty mesh! Are you sure that "
@@ -236,9 +238,9 @@ def pjit(fun: Callable,
     # TODO(yashkatariya): This is a hack. This should go away when avals have
     # is_global attribute.
     in_positional_semantics = tuple(
-        maps._PositionalSemantics.GLOBAL if isinstance(a, GDA) else maps
-        ._positional_semantics for a in args_flat)
-    out_positional_semantics = maps._positional_semantics
+        maps._PositionalSemantics.GLOBAL if isinstance(a, GDA) else maps._positional_semantics.val
+        for a in args_flat)
+    out_positional_semantics = maps._positional_semantics.val
     jaxpr, in_axis_resources_flat, out_axis_resources_flat = _pjit_jaxpr(
         flat_fun, mesh, local_in_avals, in_tree,
         hashable_pytree(in_axis_resources),
@@ -293,11 +295,60 @@ def flatten_axis_resources(what, tree, axis_resources, tupled_args):
   try:
     return tuple(flatten_axes(what, tree, axis_resources, tupled_args=tupled_args))
   except ValueError:
-    pass
+    pass  # Raise a tree prefix error below
+
+  # Tree leaves are always valid prefixes, so if there was a prefix error as
+  # assumed here, axis_resources must not be a leaf.
+  assert not treedef_is_leaf(tree_structure(axis_resources))
+
+  # Check the type directly rather than using isinstance because of namedtuples.
+  if tupled_args and (type(axis_resources) is not tuple or
+                      len(axis_resources) != len(tree.children())):
+    # We know axis_resources is meant to be a tuple corresponding to the args
+    # tuple, but while it is a non-leaf pytree, either it wasn't a tuple or it
+    # wasn't the right length.
+    msg = (f"{what} specification must be a tree prefix of the positional "
+           f"arguments tuple passed to the `pjit`-decorated function. In "
+           f"particular, {what} must either be a None, a PartitionSpec, or "
+           f"a tuple of length equal to the number of positional arguments.")
+    # If `tree` represents an args tuple, then `axis_resources` must be a tuple.
+    # TODO(mattjj,apaszke): disable implicit list casts, remove 'or list' below
+    if type(axis_resources) is not tuple:
+      msg += f" But {what} is not a tuple: got {type(axis_resources)} instead."
+    elif len(axis_resources) != len(tree.children()):
+      msg += (f" But {what} is the wrong length: got a tuple or list of length "
+              f"{len(axis_resources)} for an args tuple of length "
+              f"{len(tree.children())}.")
+
+    # As an extra hint, let's check if the user just forgot to wrap
+    # in_axis_resources in a singleton tuple.
+    if len(tree.children()) == 1:
+      try: flatten_axes(what, tree, (axis_resources,))
+      except ValueError: pass  # That's not the issue.
+      else:
+        msg += (f" Given the corresponding argument being "
+                f"passed, it looks like {what} might need to be wrapped in "
+                f"a singleton tuple.")
+
+    raise ValueError(msg)
+
   # Replace axis_resources with unparsed versions to avoid revealing internal details
-  flatten_axes(what, tree, tree_map(lambda parsed: parsed.user_spec, axis_resources),
-               tupled_args=tupled_args)
-  raise AssertionError("Please open a bug request!")  # This should be unreachable
+  axis_tree = tree_map(lambda parsed: parsed.user_spec, axis_resources)
+
+  # Because ecause we only have the `tree` treedef and not the full pytree here,
+  # we construct a dummy tree to compare against. Revise this in callers?
+  dummy_tree = tree_unflatten(tree, [PytreeLeaf()] * tree.num_leaves)
+  errors = prefix_errors(axis_tree, dummy_tree)
+  if errors:
+    e = errors[0]  # Only show information about the first disagreement found.
+    raise e(what)
+
+  # At this point we've failed to find a tree prefix error.
+  assert False, "Please open a bug report!"  # This should be unreachable.
+
+class PytreeLeaf:
+  def __repr__(self): return "pytree leaf"
+
 
 @lu.cache
 def _pjit_jaxpr(fun, mesh, local_in_avals,
@@ -327,29 +378,31 @@ def _pjit_jaxpr(fun, mesh, local_in_avals,
     # first one is a valid spec for a scalar value, while the second is not!
     _check_shapes_against_resources(
         "pjit arguments", mesh.is_multi_process, mesh.shape, local_in_avals,
-        in_axis_resources_flat)
+        in_axis_resources_flat, allow_uneven_sharding=False)
   else:
     _check_shapes_against_resources("pjit arguments", False, mesh.local_mesh.shape,
-                                    local_in_avals, in_axis_resources_flat)
+                                    local_in_avals, in_axis_resources_flat,
+                                    allow_uneven_sharding=False)
 
   global_in_avals = local_to_global(in_positional_semantics, mesh,
                                     local_in_avals, canonicalized_in_axis_resources_flat)
 
-  prev_positional = maps._positional_semantics
+  prev_positional_val = maps._positional_semantics.val
   try:
-    maps._positional_semantics = maps._PositionalSemantics.GLOBAL
+    maps._positional_semantics.val = maps._PositionalSemantics.GLOBAL
     with dispatch.log_elapsed_time(f"Finished tracing + transforming {fun.__name__} "
                                    "for pjit in {elapsed_time} sec"):
       jaxpr, global_out_avals, consts = pe.trace_to_jaxpr_dynamic(fun, global_in_avals)
   finally:
-    maps._positional_semantics = prev_positional
+    maps._positional_semantics.val = prev_positional_val
   jaxpr = core.ClosedJaxpr(jaxpr, consts)
 
   out_axis_resources_flat = flatten_axis_resources(
         "pjit out_axis_resources", out_tree(),
         out_axis_resources_thunk(), tupled_args=False)
   _check_shapes_against_resources("pjit outputs", mesh.is_multi_process, mesh.shape,
-                                  global_out_avals, out_axis_resources_flat)
+                                  global_out_avals, out_axis_resources_flat,
+                                  allow_uneven_sharding=False)
   canonicalized_out_axis_resources_flat = tree_map(_create_cpspec, out_axis_resources_flat)
   # lu.cache needs to be able to create weakrefs to outputs, so we can't return a plain tuple
   return _ListWithW([jaxpr, canonicalized_in_axis_resources_flat,
@@ -472,8 +525,7 @@ class CanonicalizedParsedPartitionSpec(ParsedPartitionSpec):
 def _prepare_axis_resources(axis_resources,
                             arg_name,
                             allow_unconstrained_dims=False):
-  # PyTrees don't treat None values as leaves, so we explicitly need
-  # to explicitly declare them as such
+  # PyTrees don't treat None values as leaves, so we use an is_leaf function.
   entries, treedef = tree_flatten(axis_resources, is_leaf=lambda x: x is None)
   what = f"{arg_name} leaf specifications"
   entries = [
@@ -501,10 +553,11 @@ def _check_unique_resources(axis_resources, arg_name):
       if multiple_uses:
         raise ValueError(f"A single {arg_name} specification can map every mesh axis "
                          f"to at most one positional dimension, but {arg_axis_resources.user_spec} "
-                         f"has duplicate entries for {maps.show_axes(multiple_uses)}")
+                         f"has duplicate entries for {pxla.show_axes(multiple_uses)}")
 
-def _check_shapes_against_resources(what: str, is_global_shape: bool, mesh_shape,
-                                    flat_avals, flat_axis_resources):
+def _check_shapes_against_resources(what: str, is_global_shape: bool,
+                                    mesh_shape, flat_avals, flat_axis_resources,
+                                    allow_uneven_sharding: bool):
   global_str = " global" if is_global_shape else ""
   for aval, aval_axis_resources in zip(flat_avals, flat_axis_resources):
     if _is_from_gda(aval_axis_resources):
@@ -524,7 +577,7 @@ def _check_shapes_against_resources(what: str, is_global_shape: bool, mesh_shape
         raise ValueError(f"One of {what} was given the resource assignment "
                          f"of {aval_axis_resources.user_spec}, but resource axis "
                          f"{e.args[0]} is undefined. Did you forget to declare the mesh?") from None
-      if shape[i] % size != 0:
+      if not allow_uneven_sharding and shape[i] % size != 0:
         raise ValueError(f"One of {what} was given the resource assignment "
                          f"of {aval_axis_resources.user_spec}, which implies that "
                          f"the{global_str} size of its dimension {i} should be "
@@ -568,12 +621,12 @@ def _pjit_lower(
   f = core.jaxpr_as_fun(jaxpr)
   f.__name__ = name
   fun = lu.wrap_init(f)
-  in_is_gda = [True if ips == maps._PositionalSemantics.GLOBAL else False
+  in_is_gda = [ips == maps._PositionalSemantics.GLOBAL
                for ips in in_positional_semantics]
   return pxla.lower_mesh_computation(
       fun, name, resource_env.physical_mesh,
       in_axes, out_axes, donated_invars,
-      True, jaxpr.in_avals, tile_by_mesh_axes=False, in_is_gda=in_is_gda)
+      True, jaxpr.in_avals, tiling_method=None, in_is_gda=in_is_gda)
 
 
 def _pjit_abstract_eval(*args, jaxpr, out_axis_resources, resource_env,
@@ -587,6 +640,7 @@ def _pjit_translation_rule(ctx, avals_in, avals_out, *in_nodes, name,
                            jaxpr, in_axis_resources, out_axis_resources,
                            resource_env, donated_invars, in_positional_semantics,
                            out_positional_semantics):
+  # TODO: Make this into an MLIR rule and use manual_axes!
   mesh = resource_env.physical_mesh
   subc = xc.XlaBuilder(f"pjit_{name}")
 
@@ -846,7 +900,7 @@ def _check_resources_against_named_axes(what, aval, pos_axis_resources, named_ax
         f"{pos_axis_resources.unsynced_user_spec(SpecSync.DIM_PERMUTE)} "
         f"that uses one or more mesh axes already used by xmap to partition "
         f"a named axis appearing in its named_shape (both use mesh axes "
-        f"{maps.show_axes(overlap)})")
+        f"{pxla.show_axes(overlap)})")
 
 def _resource_typing_pjit(avals, params, source_info, resource_env, named_axis_resources):
   jaxpr = params["jaxpr"]
@@ -874,12 +928,12 @@ def with_sharding_constraint(x, axis_resources):
   axis_resources_flat = tuple(
       flatten_axes("with_sharding_constraint axis_resources",
                    tree, parsed_axis_resources))
-  resource_env = maps.thread_resources.env
+  resource_env = pxla.thread_resources.env
   mesh = resource_env.physical_mesh
   _check_shapes_against_resources(
       "with_sharding_constraint arguments",
       mesh.is_multi_process, mesh.shape,
-      x_flat, axis_resources_flat)
+      x_flat, axis_resources_flat, allow_uneven_sharding=True)
   outs = [sharding_constraint_p.bind(y, axis_resources=r, resource_env=resource_env)
           for y, r in safe_zip(x_flat, axis_resources_flat)]
   return tree_unflatten(tree, outs)
@@ -906,7 +960,8 @@ def _sharding_constraint_translation_rule(ctx, avals_in, avals_out, x_node, *,
       xla.set_sharding_proto(
           ctx.builder,
           x_node,
-          get_aval_sharding_proto(aval, axis_resources, mesh),
+          get_aval_sharding_proto(
+              aval, axis_resources, mesh, allow_uneven_axes=True),
           unspecified_dims=get_unconstrained_dims(axis_resources))
   ]
 xla.register_translation(sharding_constraint_p, _sharding_constraint_translation_rule)
@@ -918,7 +973,12 @@ def _sharding_constraint_mhlo_lowering(ctx, x_node, *, axis_resources,
   return [
       mlir.wrap_with_sharding_op(
           x_node,
-          get_aval_sharding_proto(aval, axis_resources, mesh),
+          get_aval_sharding_proto(
+              aval,
+              axis_resources,
+              mesh,
+              ctx.module_context.axis_context,
+              allow_uneven_axes=True),
           unspecified_dims=get_unconstrained_dims(axis_resources))
   ]
 mlir.register_lowering(sharding_constraint_p,
@@ -957,11 +1017,18 @@ def get_array_mapping(axis_resources: ParsedPartitionSpec) -> pxla.ArrayMapping:
 
 def get_aval_sharding_proto(aval: core.AbstractValue,
                             axis_resources: ParsedPartitionSpec,
-                            mesh: maps.Mesh) -> xc.OpSharding:
+                            mesh: maps.Mesh,
+                            axis_ctx: Optional[mlir.SPMDAxisContext] = None,
+                            allow_uneven_axes: bool = False) -> xc.OpSharding:
   array_mapping = get_array_mapping(axis_resources)
-  sharding_spec = pxla.mesh_sharding_specs(mesh.shape, mesh.axis_names)(
-      aval, array_mapping)
-  return sharding_spec.sharding_proto()
+  sharding_spec = pxla.mesh_sharding_specs(
+      mesh.shape, mesh.axis_names, allow_uneven_axes=True)(aval, array_mapping)
+  special_axes = {}
+  if axis_ctx is not None:
+    axis_names = mesh.axis_names
+    for manual_axis in axis_ctx.manual_axes:
+      special_axes[axis_names.index(manual_axis)] = xc.OpSharding.Type.MANUAL
+  return sharding_spec.sharding_proto(special_axes=special_axes)
 
 
 def get_unconstrained_dims(axis_resources: ParsedPartitionSpec):
@@ -1135,7 +1202,13 @@ def parse_op_sharding(op_sharding, mesh):
         dim_size //= axis_size
         dim_partitions.append(axis)
       partitions.append(tuple(dim_partitions))
-    if op_sharding.replicate_on_last_tile_dim:
+    if op_sharding.last_tile_dims == [xc.OpSharding.Type.REPLICATED]:
+      replicate_on_last_tile_dim = True
+    else:
+      replicate_on_last_tile_dim = op_sharding.replicate_on_last_tile_dim
+      if op_sharding.last_tile_dims:
+        raise NotImplementedError("Unhandled OpSharding type. Please open a bug report!")
+    if replicate_on_last_tile_dim:
       partitions = partitions[:-1]
     return ParsedPartitionSpec('<internally generated spec>', partitions)
   else:
