@@ -24,7 +24,7 @@ import itertools
 import re
 import typing
 from typing import (Any, Callable, Dict, List, Optional, Sequence, Set, Tuple,
-                    Type, Union)
+                    Type, Union, FrozenSet)
 from typing_extensions import Protocol
 import warnings
 
@@ -39,7 +39,6 @@ from jax._src.lib.mlir.dialects import chlo
 from jax._src.lib.mlir.dialects import mhlo
 from jax._src.lib.mlir.dialects import std
 from jax._src.lib import xla_client as xc
-import jax._src.pretty_printer as pp
 from jax._src import source_info_util
 import jax._src.util as util
 import jax.interpreters.ad as ad
@@ -259,6 +258,8 @@ def _device_array_constant_handler(val, canonicalize_types):
 for t in device_array.device_array_types:
   register_constant_handler(t, _device_array_constant_handler)
 
+register_constant_handler(
+  core.Token, lambda _, __: [mhlo.CreateTokenOp(mhlo.TokenType.get())])
 
 # Source locations
 
@@ -266,9 +267,7 @@ def _source_info_to_location(
     primitive: core.Primitive, params: Dict,
     source_info: source_info_util.SourceInfo,
     name_stack: str = "") -> ir.Location:
-  eqn_str = str(
-      pp.text(name_stack) +
-      core.pp_eqn_compact(primitive.name, params, core.JaxprPpContext()))
+  eqn_str = name_stack + core.str_eqn_compact(primitive.name, params)
   frame = source_info_util.user_frame(source_info)
   if frame is None:
     loc = ir.Location.unknown()
@@ -290,6 +289,42 @@ def make_ir_context() -> ir.Context:
   return context
 
 
+Mesh = Any
+MeshAxisName = Any
+
+@dataclasses.dataclass(frozen=True)
+class SPMDAxisContext:
+  """A hardware axis context for parallel computations that use the GSPMD partitioner.
+
+  This includes the mesh that will later by used to execute this computation,
+  as well as a set of mesh axes that are currently (e.g. because the current lowering
+  is invoked inside an xmap) lowered in the MANUAL sharding mode.
+  """
+  mesh: Mesh
+  manual_axes: FrozenSet[MeshAxisName] = frozenset()
+
+  @property
+  def axis_env(self):
+    # TODO: Should we restrict the mesh to manual axes? This depends on the
+    # semantics of ReplicaId in partially manual computations, but I don't
+    # know what they are...
+    return xla.AxisEnv(nreps=1, names=(), sizes=())
+
+  def extend_manual(self, axes: FrozenSet[MeshAxisName]) -> 'SPMDAxisContext':
+    return SPMDAxisContext(self.mesh, self.manual_axes | axes)
+
+
+@dataclasses.dataclass(frozen=True)
+class ReplicaAxisContext:
+  """A hardware axis context for parallel computations that are partitioned by JAX.
+
+  Unlike in the SPMDAxisContext, this means that JAX might need to emit calls to
+  explicit collectives.
+  """
+  axis_env: xla.AxisEnv
+
+AxisContext = Union[SPMDAxisContext, ReplicaAxisContext]
+
 @dataclasses.dataclass
 class ModuleContext:
   """Module-wide context information for MLIR lowering."""
@@ -298,14 +333,18 @@ class ModuleContext:
   ip: ir.InsertionPoint
   symbol_table: ir.SymbolTable
   platform: str
-  axis_env: xla.AxisEnv
+  axis_context: AxisContext
   name_stack: str
 
   # Cached primitive lowerings.
   cached_primitive_lowerings: Dict[Any, builtin.FuncOp]
 
+  @property
+  def axis_env(self) -> xla.AxisEnv:
+    return self.axis_context.axis_env
+
   def __init__(
-      self, platform: str, axis_env: xla.AxisEnv, name_stack: str,
+      self, platform: str, axis_context: AxisContext, name_stack: str,
       context: Optional[ir.Context] = None,
       module: Optional[ir.Module] = None,
       ip: Optional[ir.InsertionPoint] = None,
@@ -317,7 +356,7 @@ class ModuleContext:
     self.ip = ip or ir.InsertionPoint(self.module.operation.opview.body)
     self.symbol_table = symbol_table or ir.SymbolTable(self.module.operation)
     self.platform = platform
-    self.axis_env = axis_env
+    self.axis_context = axis_context
     self.name_stack = name_stack
     self.cached_primitive_lowerings = ({} if cached_primitive_lowerings is None
                                        else cached_primitive_lowerings)
@@ -372,7 +411,7 @@ _module_name_regex = re.compile(r"[^\w.-]")
 
 def lower_jaxpr_to_module(
     module_name: str, jaxpr: core.ClosedJaxpr, platform: str,
-    axis_env: xla.AxisEnv,
+    axis_context: AxisContext,
     name_stack: str, donated_args: Sequence[bool],
     replicated_args: Optional[Sequence[bool]] = None,
     arg_shardings: Optional[Sequence[Optional[xc.OpSharding]]] = None,
@@ -396,7 +435,7 @@ def lower_jaxpr_to_module(
       msg = f"Donation is not implemented for {platform}.\n{msg}"
     warnings.warn(f"Some donated buffers were not usable: {', '.join(unused_donations)}.\n{msg}")
 
-  ctx = ModuleContext(platform, axis_env, name_stack)
+  ctx = ModuleContext(platform, axis_context, name_stack)
   with ctx.context, ir.Location.unknown(ctx.context):
     # Remove module name characters that XLA would alter. This ensures that
     # XLA computation preserves the module name.
@@ -423,6 +462,11 @@ def module_to_string(module: ir.Module) -> str:
 
 def _set_up_aliases(avals_in, avals_out, donated_args):
   input_output_aliases = [None] * len(avals_in)
+  # To match-up in-avals to out-avals we only care about the number of
+  # bytes, so we strip off unrelated aval metadata (eg. the named shape)
+  strip_metadata = lambda a: a.strip_named_shape().strip_weak_type()
+  avals_in = map(strip_metadata, avals_in)
+  avals_out = map(strip_metadata, avals_out)
 
   donations = collections.defaultdict(collections.deque)
   for i, (aval, donated) in enumerate(zip(avals_in, donated_args)):
@@ -769,10 +813,11 @@ def convert_mhlo(x, aval_in, aval_out):
         ir.StringAttr.get("NE"), ir.StringAttr.get(compare_type)).result
   return mhlo.ConvertOp(aval_to_ir_type(aval_out), x).result
 
-
-def wrap_with_sharding_op(x,
-                          sharding_proto: xc.OpSharding,
-                          unspecified_dims: Optional[Set[int]] = None):
+def _wrap_with_spmd_op(name: str,
+                       result_type: ir.Type,
+                       x: ir.Value,
+                       sharding_proto: xc.OpSharding,
+                       unspecified_dims: Optional[Set[int]] = None):
   # unspecified_dims indicate dimensions whose shardings are not specified and
   # XLA sharding propagation can change them.
   if unspecified_dims:
@@ -780,8 +825,8 @@ def wrap_with_sharding_op(x,
         [str(i) for i in sorted(unspecified_dims)]) + "]"
   else:
     backend_config = ""
-  op = mhlo.CustomCallOp([x.type], [x],
-                         call_target_name=ir.StringAttr.get("Sharding"),
+  op = mhlo.CustomCallOp([result_type], [x],
+                         call_target_name=ir.StringAttr.get(name),
                          has_side_effect=ir.BoolAttr.get(False),
                          backend_config=ir.StringAttr.get(backend_config),
                          api_version=i32_attr(1),
@@ -792,6 +837,14 @@ def wrap_with_sharding_op(x,
       sharding_proto.SerializeToString())
   return op.result
 
+def wrap_with_sharding_op(x: ir.Value,
+                          sharding_proto: xc.OpSharding,
+                          unspecified_dims: Optional[Set[int]] = None):
+  return _wrap_with_spmd_op("Sharding", x.type, x, sharding_proto, unspecified_dims)
+
+wrap_with_full_to_shard_op = partial(_wrap_with_spmd_op, "SPMDFullToShardShape")
+wrap_with_shard_to_full_op = partial(_wrap_with_spmd_op, "SPMDShardToFullShape")
+
 def set_sharding(op, sharding_proto: xc.OpSharding):
   op.attributes["mhlo.sharding"] = ir.StringAttr.get(
       sharding_proto.SerializeToString())
