@@ -204,7 +204,8 @@ from jax._src.lax.lax import (
   rsqrt as rsqrt,
   rsqrt_p as rsqrt_p,
   select as select,
-  select_p as select_p,
+  select_n as select_n,
+  select_n_p as select_n_p,
   shift_left as shift_left,
   shift_left_p as shift_left_p,
   shift_right_arithmetic as shift_right_arithmetic,
@@ -364,6 +365,7 @@ from jax._src.lax.parallel import (
   xeinsum as xeinsum,
 )
 from jax._src.lax.other import (
+  conv_general_dilated_local as conv_general_dilated_local,
   conv_general_dilated_patches as conv_general_dilated_patches
 )
 from jax._src.ad_util import stop_gradient_p as stop_gradient_p
